unit SharedDynamicArrays;

interface
Type TSharedArray=class
private
FBaseOffset:cardinal; //Starting array's offset in shared memory
FSizeOfElem:cardinal; //Size of one record in the array
FHandle:THandle; //CreateFileMapping
FShareAddress:pointer; //MapViewOfFile
FShareName:PAnsiChar;
FOldLength:cardinal;
Function GetLength:cardinal;
Procedure SetLength(const NewLength:cardinal);
function Get(i:cardinal):pointer;
procedure Open;
public
Success:boolean;
Procedure Clear;
Procedure ZeroElement(const i:cardinal);
Constructor Create(const ShareName:PAnsiChar; AlreadyExists:boolean; SizeOfElem:cardinal;
InitialLength:cardinal=0; InitialOffset:cardinal=0);
Destructor Destroy; override;
property Length:cardinal read GetLength write SetLength;
property Index[i:cardinal]:pointer read get; default;
       end;
implementation
uses
{$ifdef debug}sysutils,{$endif} logging, windows;

{ TSharedArray }

procedure TSharedArray.Clear;
begin
{$ifdef debug}
log('TSharedArray.Clear()');
{$endif}
Success:=false;
if FShareAddress=nil then
{$ifdef debug}
begin
log('FShareAddress = nil');
{$endif}
exit;
{$ifdef debug}
end;
{$endif}
if length > FOldLength then begin
{$ifdef debug}
log(format('length(%d) > FOldLength(%d)', [Length, FOldLength]));
{$endif}
open;
if not success then exit;
end;
zeromemory(pointer(cardinal(FShareAddress) + FBaseOffset +4), FSizeOfElem * FOldLength);
success:=true;
end;

constructor TSharedArray.Create(const ShareName: PAnsiChar; AlreadyExists: boolean;
  SizeOfElem, InitialLength, InitialOffset: cardinal);
begin
{$ifdef debug}
log(format('TSharedArray.Create(%s, %d, %d, %d, %d)', [ShareName, integer(AlreadyExists), SizeOfElem, InitialLength, InitialOffset]));
{$endif}
Success:=false;
if ((not AlreadyExists) and (InitialLength = 0)) or (SizeOfElem = 0) then exit;
FBaseOffset:=InitialOffset;
FSizeOfElem:=SizeOfElem;
FShareName:=ShareName;
if AlreadyExists then Open
else
Length:=InitialLength;
end;
destructor TSharedArray.Destroy;
begin
{$ifdef debug}
log('TSharedArray.Destroy()');
{$endif}
              UnMapViewOfFile(FShareAddress);
//CloseHandle(FHandle);
  inherited;
end;

function TSharedArray.Get(i: cardinal): pointer;
begin
{$ifdef debug}
log(format('TSharedArray.Get(%d)', [i]));
{$endif}
result:=nil;
Success:=false;
if (FShareAddress = nil) then exit;
if length > FOldLength then begin
{$ifdef debug}
log(format('length(%d) > FOldLength(%d)', [Length, FOldLength]));
{$endif}
open;
if not success then exit;
end;
if (i > Length) then
              {$ifdef debug}
begin
log('i > Length');
{$endif}
exit;
{$ifdef debug}
end;
{$endif}
    result:=pointer(cardinal(FShareAddress) + FBaseOffset + 4 + FSizeOfElem * (i - 1)); //index must be greater than 0!
Success:=true;
end;

function TSharedArray.GetLength: cardinal;
begin
{$ifdef debug}
log('TSharedArray.GetLength()');
{$endif}
Success:=false;
result:=0;
if FShareAddress=nil then
{$ifdef debug}
begin
log('FShareAddress = nil');
{$endif}
exit;
{$ifdef debug}
end;
{$endif}
result:=cardinal(pointer(cardinal(FShareAddress) + FBaseOffset)^);
Success:=true;
{$ifdef debug}
log(format('length = %d', [result]));
{$endif}
end;

procedure TSharedArray.Open;
begin
{$ifdef debug}
log('TSharedArray.Open');
{$endif}
success:=false;
FHandle:=OpenFileMapping(FILE_MAP_ALL_ACCESS, false, FShareName);
                                                               if FHandle = 0 then begin
log('TSharedArray.Open: OpenFileMapping failed: ' + syserrormessage(getlasterror));
exit;
end;
FShareAddress:=MapViewOfFile(FHandle, FILE_MAP_ALL_ACCESS, 0, FBaseOffset, 0);
if FShareAddress=nil then begin
log('TSharedArray.Open: MapViewOfFile failed: ' + syserrormessage(getlasterror));
exit;
end;
FOldLength:=length;
Success:=true;
end;

procedure TSharedArray.SetLength(const NewLength: cardinal);
var TempBuf:pointer;
begin
{$ifdef debug}
log(format('TSharedArray.SetLength(%d)', [NewLength]));
{$endif}
Success:=false;
TempBuf:=nil;
if (FShareAddress <> nil) and (Length < NewLength) then begin
{$ifdef debug}
log('Moving old data');
{$endif}
getmem(TempBuf, FBaseOffset + 4 + length * FSizeOfElem);
move(FShareAddress^, TempBuf^, FBaseOffset +4 + Length * FSizeOfElem);
FOldLength:=Length;
UnMapViewOfFile(FShareAddress);
CloseHandle(FHandle);
end;
FHandle:=CreateFileMapping($FFFFFFFF, nil, PAGE_READWRITE, 0,
FBaseOffset +4 +NewLength * FSizeOfElem, FShareName);
                                                               if FHandle = 0 then begin
log('TSharedArray.SetLength: CreateFileMapping failed: ' + syserrormessage(getlasterror));
exit;
end;
FShareAddress:=MapViewOfFile(FHandle, FILE_MAP_ALL_ACCESS, 0, FBaseOffset, 0);
if FShareAddress=nil then begin
log('TSharedArray.SetLength: MapViewOfFile failed: ' + syserrormessage(getlasterror));
exit;
end;
if TempBuf<>nil then begin
move(TempBuf^, FShareAddress^, FBaseOffset +4 + FOldLength * FSizeOfElem);
freemem(TempBuf);
end;
cardinal(pointer(cardinal(FShareAddress) + FBaseOffset)^):=NewLength;
FOldLength:=length;
Success:=true;
end;

procedure TSharedArray.ZeroElement(const i: cardinal);
begin
{$ifdef debug}
log(format('TSharedArray.ZeroElement(%d)', [i]));
{$endif}
zeromemory(self[i], FSizeOfElem);
end;

end.
