//summary: saving and manipulation hooked portions of text
unit storage;

interface
uses windows;

type TGDIHookRec = packed record
x,y:integer;
handle:THandle;
text:PWideChar;
end;

const
MaxChars=1024; //maximum characters for portion

Type TPortion=packed record
rect:TRect;
handle:THandle;
str:Array[1..MaxChars] of WideChar;
end;
Function InitStorage:boolean;
function StorageText(const text:PWideChar; left,top,right,bottom:integer; handle:THandle):boolean; overload;
function StorageText(const text:PAnsiChar; left,top,right,bottom:integer; handle:THandle):boolean; overload;
Function OpenStorage:boolean;
Procedure ClearStorage;
Procedure CloseStorage;
Function GetTextAtCoords(x,y:integer; var TextBuf:pointer; var elems:cardinal):integer; stdcall;
Function GetTextInRect(var rect:TRect; var TextBuf:pointer; var Elems:DWORD):integer; stdcall;
Procedure ClearStorageInRect(const left,top,right,bottom:integer); stdcall;
function GetTextFromWindow(const handle:THandle; var TextBuf:pointer; var elems:DWORD):Integer; stdcall;
Function PTInRect(rect:TRect; PT:TPoint):boolean;
implementation
uses {$ifdef debug}sysutils, {$endif}logging, SharedDynamicArrays;
const StorageShareName='NVDAGDIHookDLL_Share'; //identifier for CreateFileMapping
ShareMutexName = StorageShareName+'_mutex';
//errors
GDIHook_ok = 0;
GDIHook_NotFound = -1;
GDIHook_InternalError = -2;
var ShareMutex:THandle=0;
Buffer:TSharedArray=nil;
Procedure ClearStorage;
begin
{$ifdef debug}
log('ClearStorage called');
{$endif}
if (Buffer=nil) or (ShareMutex=0) then exit;
              {$ifdef debug}
log('Waiting for access');
{$endif}
WaitForSingleObject(sharemutex, infinite);
try
Buffer.Clear;
except
log('Buffer.Clear() failed');
end;
ReleaseMutex(ShareMutex);
{$ifdef debug}
log('Mutex released');
{$endif}
end;

Function InitStorage:boolean;
begin
{$ifdef debug}
log('InitStorage called');
{$endif}
result:=false;
ShareMutex:=CreateMutex(nil, false, ShareMutexName);
if ShareMutex = 0 then begin
log('InitStorage: CreateMutex failed');
exit;
end;
WaitForSingleObject(ShareMutex, infinite);
try
Buffer:=TSharedArray.Create(StorageShareName, false, sizeof(TPortion), 100);
except
log('TSharedArray.create failed');
end;
ReleaseMutex(ShareMutex);
if not buffer.Success then exit;
result:=true;
end;

function StorageText(const text:PWideChar; left,top,right,bottom:integer; handle:THandle):boolean; overload;
var c:cardinal;
r,r1:TRect;
//i:integer;
begin
{$ifdef debug}
log('StorageText: called with following args:');
log(format('Text: "%s", left: %d, top: %d, right: %d, bottom: %d, Handle: %d',
[widechartostring(text), left, top, right, bottom, handle]));
{$endif}
result:=false;
if buffer=nil then exit;
if (left = 0) and (top = 0) and (right = 0) and (bottom = 0) then exit;
if lstrlenW(text) = 0 then exit;
{i:=pos(#13#10, widechartostring(text));
if i > 0 then lstrcpynW(text, text, i);
i:=pos(#13, widechartostring(text));
if i > 0 then lstrcpynW(text, text, i);
i:=pos(#10, widechartostring(text));
if i > 0 then lstrcpynW(text, text, i);}
setrect(r, left, top, right, bottom);
{$ifdef debug}
log('Waiting for access');
{$endif}
WaitForSingleObject(ShareMutex, infinite);
try
for c:=1 to Buffer.Length do begin
if TPortion(Buffer[c]^).str[1]=#0 then continue
else
if (intersectrect(r1, r, TPortion(buffer[c]^).rect))
and (TPortion(buffer[c]^).handle = handle) then begin
{$ifdef debug}
log(format('StorageText: record near that coords allready exists ' +
'(with text "%s"), replacing with passed text',
[widechartostring(@TPortion(buffer[c]^).str[1])]));
{$endif}
lstrcpyW(@TPortion(Buffer[c]^).str[1], text);
copyrect(TPortion(Buffer[c]^).rect, r);
ReleaseMutex(ShareMutex);
{$ifdef debug}
log('Mutex Released');
{$endif}
result:=true;
exit;
end;
end;
except
log('Exception while look over');
end;
try
for c:=1 to Buffer.Length do
if TPortion(Buffer[c]^).str[1] = #0 then begin
{$ifdef debug}
log(format('StorageText: adding new record to position %d', [c]));
{$endif}
lstrcpyW(@TPortion(Buffer[c]^).str[1], text);
copyrect(TPortion(Buffer[c]^).rect, r);
TPortion(Buffer[c]^).handle:=handle;
ReleaseMutex(ShareMutex);
{$ifdef debug}
log('Mutex released');
{$endif}
result:=true;
exit;
end;
except
log('Exception while finding empty record');
end;
{$ifdef debug}
log('empty records not found');
{$endif}
//empty records not found, so array must be resized
c:=Buffer.Length +1;
{$ifdef debug}
log(format('Current length: %d', [buffer.Length]));
{$endif}
Buffer.Length := Buffer.Length * 2;
{$ifdef debug}
log(format('after resizing: %d', [buffer.length]));
{$endif}
lstrcpyW(@TPortion(Buffer[c]^).str[1], text);
copyrect(TPortion(Buffer[c]^).rect, r);
TPortion(Buffer[c]^).handle:=handle;
ReleaseMutex(ShareMutex);
{$ifdef debug}
log('Mutex released');
{$endif}
result:=true;
end;

Procedure CloseStorage;
begin
{$ifdef debug}
log('CloseStorage called');
{$endif}
try
Buffer.Free;
Buffer:=nil;
except
log('exception while calling buffer.free');
end;
end;

Function GetTextAtCoords(x,y:integer; var TextBuf:pointer; var elems:cardinal):integer; stdcall;
var
c,d:cardinal;
tempbuf:array of cardinal;
pt:TPoint;
begin
{$ifdef debug}
log('GetTextAtCoords: called with following args:');
log(format('X: %d, Y: %d', [x, y]));
{$endif}
result:=GDIHook_NotFound;
if (Buffer=nil) then
if not OpenStorage then begin
log('GetTextAtCoords failed: Storage not initialized');
result:=GDIHook_InternalError;
exit;
end;
pt.X:=x;
pt.Y:=y;
d:=0;
setlength(tempbuf, Buffer.Length+1);
{$ifdef debug}
log('Waiting for access');
{$endif}
WaitForSingleObject(ShareMutex, infinite);
try
    for c:=1 to Buffer.Length do
    if (TPortion(Buffer[c]^).str[1] > #0) and (ptinrect(TPortion(Buffer[c]^).rect, pt)) then begin
{$ifdef debug}
log(format('GetTextAtCoords: found text "%s" (left: %d, top: %d, right: %d, bottom: %d)',
[widechartostring(@TPortion(Buffer[c]^).str[1]), TPortion(Buffer[c]^).rect.Left,
TPortion(Buffer[c]^).rect.Top, TPortion(Buffer[c]^).rect.Right, TPortion(BufFer[c]^).rect.Bottom]));
{$endif}
inc(d);
tempBuf[d]:=c;
end;
except
log('Exception while look over');
end;
if d = 0 then begin
releaseMutex(ShareMutex);
{$ifdef debug}
log('Mutex released');
{$endif}
exit;
end;
c:=1;
try
while c < d do begin
if (TPortion(Buffer[tempbuf[c]]^).rect.Top > TPortion(Buffer[tempbuf[c+1]]^).rect.top)
or ((TPortion(Buffer[tempbuf[c]]^).rect.Top = TPortion(Buffer[tempbuf[c+1]]^).rect.top)
and (TPortion(Buffer[tempbuf[c]]^).rect.Left > TPortion(Buffer[tempbuf[c+1]]^).rect.left))
then begin
inc(tempbuf[c+1], tempbuf[c]);
tempbuf[c] := tempbuf[c+1] - tempbuf[c];
tempbuf[c+1]:=tempbuf[c+1] - tempbuf[c];
c:=1;
continue;
end;
inc(c);
end;
except
log('exception while sorting');
end;
getmem(textbuf, sizeof(TGDIHookRec) * d);
try
for c:= 1 to d do begin
TGDIHookRec(pointer(dword(TextBuf) + sizeof(TGDIHookRec) * (c - 1))^).x:=
TPortion(Buffer[tempbuf[c]]^).rect.Left;
TGDIHookRec(pointer(dword(TextBuf) + sizeof(TGDIHookRec) * (c - 1))^).y:=
TPortion(Buffer[tempbuf[c]]^).rect.Top;
TGDIHookRec(pointer(dword(TextBuf) + sizeof(TGDIHookRec) * (c - 1))^).handle:=
TPortion(Buffer[tempbuf[c]]^).handle;
getmem(TGDIHookRec(pointer(dword(TextBuf) + sizeof(TGDIHookRec) * (c - 1))^).text,
lstrlenW(@TPortion(Buffer[tempbuf[c]]^).str[1]) * 2 +2);
lstrcpyW(TGDIHookRec(pointer(dword(TextBuf) + sizeof(TGDIHookRec) * (c - 1))^).text,
@TPortion(Buffer[tempbuf[c]]^).str[1]);
end;
except
log('Exception while preparing data');
end;
ReleaseMutex(ShareMutex);
{$ifdef debug}
log('Mutex released');
{$endif}
Elems := d;
{$ifdef debug}
log(format('GetTextAtCoords: elems = %d', [elems]));
{$endif}
result:=GDIHook_OK;
end;

Function GetTextInRect(var rect:TRect; var TextBuf:pointer; var Elems:DWORD):integer; stdcall;
var
c,d:cardinal;
TempBuf:array of cardinal;
r:TRect;
begin
{$ifdef debug}
log('GetTextInRect: called with following args:');
log(format('rect.left: %d, rect.top: %d, rect.right: %d, rect.bottom: %d', [rect.Left, rect.Top, rect.Right, rect.Bottom]));
{$endif}
result:=GDIHook_NotFound;
if (Buffer=nil)then
if not OpenStorage then begin
log('GetTextInRect failed: Storage not initialized');
result:=GDIHook_InternalError;
exit;
end;
d:=0;
setlength(tempbuf, Buffer.Length+1);
{$ifdef debug}
log('Waiting for access');
{$endif}
WaitForSingleObject(ShareMutex, infinite);
try
    for c:=1 to Buffer.Length do
    if (TPortion(Buffer[c]^).str[1] > #0)
and (intersectrect(r, rect, TPortion(Buffer[c]^).rect)) then begin
{$ifdef debug}
log(format('GetTextInRect: found text "%s" (left: %d, top: %d, right: %d, bottom: %d)',
[widechartostring(@TPortion(Buffer[c]^).str[1]), TPortion(Buffer[c]^).rect.Left,
TPortion(Buffer[c]^).rect.Top, TPortion(Buffer[c]^).rect.Right, TPortion(BufFer[c]^).rect.Bottom]));
{$endif}
inc(d);
tempBuf[d]:=c;
end;
except
log('Exception while look over');
end;
if d = 0 then begin
releaseMutex(ShareMutex);
{$ifdef debug}
log('Mutex released');
{$endif}
exit;
end;
c:=1;
try
while c < d do begin
if (TPortion(Buffer[tempbuf[c]]^).rect.Top > TPortion(Buffer[tempbuf[c+1]]^).rect.top)
or ((TPortion(Buffer[tempbuf[c]]^).rect.Top = TPortion(Buffer[tempbuf[c+1]]^).rect.top)
and (TPortion(Buffer[tempbuf[c]]^).rect.Left > TPortion(Buffer[tempbuf[c+1]]^).rect.left))
then begin
inc(tempbuf[c+1], tempbuf[c]);
tempbuf[c] := tempbuf[c+1] - tempbuf[c];
tempbuf[c+1]:=tempbuf[c+1] - tempbuf[c];
c:=1;
continue;
end;
inc(c);
end;
except
log('exception while sorting');
end;
getmem(textbuf, sizeof(TGDIHookRec) * d);
try
for c:= 1 to d do begin
TGDIHookRec(pointer(dword(TextBuf) + sizeof(TGDIHookRec) * (c - 1))^).x:=
TPortion(Buffer[tempbuf[c]]^).rect.Left;
TGDIHookRec(pointer(dword(TextBuf) + sizeof(TGDIHookRec) * (c - 1))^).y:=
TPortion(Buffer[tempbuf[c]]^).rect.Top;
TGDIHookRec(pointer(dword(TextBuf) + sizeof(TGDIHookRec) * (c - 1))^).handle:=
TPortion(Buffer[tempbuf[c]]^).handle;
getmem(TGDIHookRec(pointer(dword(TextBuf) + sizeof(TGDIHookRec) * (c - 1))^).text,
lstrlenW(@TPortion(Buffer[tempbuf[c]]^).str[1]) * 2 +2);
lstrcpyW(TGDIHookRec(pointer(dword(TextBuf) + sizeof(TGDIHookRec) * (c - 1))^).text,
@TPortion(Buffer[tempbuf[c]]^).str[1]);
end;
except
log('exception while preparing data');
end;
ReleaseMutex(ShareMutex);
{$ifdef debug}
log('Mutex released');
{$endif}
Elems := d;
{$ifdef debug}
log(format('GetTextInRect: elems = %d', [elems]));
{$endif}
                      result:=GDIHook_OK;
                      end;

function StorageText(const text:PAnsiChar; left,top,right,bottom:integer; handle:THandle):boolean; overload;
var w:PWideChar;
l:integer;
begin
l:=lstrlen(text) * 2 +2;
getmem(w, l);
result:=StorageText(stringtowidechar(text, w, l), left, top, right, bottom, handle);
end;

Procedure ClearStorageInRect(const left,top,right,bottom:integer); stdcall;
var
c:cardinal;
pt:TPoint;
r:TRect;
begin
{$ifdef debug}
log('ClearStorageInRect: called with following args:');
log(format('left: %d, top: %d, right: %d, bottom: %d', [left,top,right,bottom]));
{$endif}
if (Buffer=nil) then
if not OpenStorage then begin
log('ClearStorageInRect failed: Storage not initialized');
exit;
end;
SetRect(r, left, top, right, bottom);
{$ifdef debug}
log('Waiting for access');
{$endif}
WaitForSingleObject(ShareMutex, 10);
try
for c:=1 to Buffer.Length do
if TPortion(Buffer[c]^).str[1]>#0 then begin
pt.X:=TPortion(Buffer[c]^).rect.Left;
pt.Y:=TPortion(Buffer[c]^).rect.Top;
if ptinrect(r, pt) then
{$ifdef debug}
begin
log(format('ClearTextInRect: found text "%s" (left: %d, top: %d, right: %d, bottom: %d)',
[widechartostring(@TPortion(Buffer[c]^).str[1]), TPortion(Buffer[c]^).rect.Left,
TPortion(Buffer[c]^).rect.Top, TPortion(Buffer[c]^).rect.Right, TPortion(Buffer[c]^).rect.Top]));
{$endif}
Buffer.ZeroElement(c);
{$ifdef debug}
end;
{$endif}
end;
except
log('exception while look over and clean');
end;
ReleaseMutex(ShareMutex);
{$ifdef debug}
log('Mutex released');
{$endif}
end;

function GetTextFromWindow(const handle:THandle; var TextBuf:pointer; var elems:DWORD):integer; stdcall;
var
c,d:cardinal;
TempBuf:array of cardinal;
begin
{$ifdef debug}
log('GetTextFromWindow: called with following args:');
log(format('handle: %d', [handle]));
{$endif}
    result:=GDIHook_NotFound;
if Buffer=nil then
if not OpenStorage then begin
log('GetTextFromWindow failed: Storage not initialized');
result:=GDIHook_InternalError;
exit;
end;
d:=0;
{$ifdef debug}
log('Waiting for access');
{$endif}
WaitForSingleObject(ShareMutex, infinite);
setlength(tempbuf, Buffer.Length+1);
try
    for c:=1 to Buffer.Length do
    if (TPortion(Buffer[c]^).str[1] > #0)
and (TPortion(Buffer[c]^).handle = handle) then begin
{$ifdef debug}
log(format('GetTextFromWindow: found text "%s" (left: %d, top: %d, right: %d, bottom: %d)',
[widechartostring(@TPortion(Buffer[c]^).str[1]), TPortion(Buffer[c]^).rect.Left,
TPortion(Buffer[c]^).rect.Top, TPortion(Buffer[c]^).rect.Right, TPortion(BufFer[c]^).rect.Bottom]));
{$endif}
inc(d);
tempBuf[d]:=c;
end;
except
log('exception while looking over');
end;
if d = 0 then begin
releaseMutex(ShareMutex);
{$ifdef debug}
log('Mutex released');
{$endif}
exit;
end;
c:=1;
try
while c < d do begin
if (TPortion(Buffer[tempbuf[c]]^).rect.Top > TPortion(Buffer[tempbuf[c+1]]^).rect.top)
or ((TPortion(Buffer[tempbuf[c]]^).rect.Top = TPortion(Buffer[tempbuf[c+1]]^).rect.top)
and (TPortion(Buffer[tempbuf[c]]^).rect.Left > TPortion(Buffer[tempbuf[c+1]]^).rect.left))
then begin
inc(tempbuf[c+1], tempbuf[c]);
tempbuf[c] := tempbuf[c+1] - tempbuf[c];
tempbuf[c+1]:=tempbuf[c+1] - tempbuf[c];
c:=1;
continue;
end;
inc(c);
end;
except
log('Exception while sorting');
end;
getmem(textbuf, sizeof(TGDIHookRec) * d);
try
for c:= 1 to d do begin
TGDIHookRec(pointer(dword(TextBuf) + sizeof(TGDIHookRec) * (c - 1))^).x:=
TPortion(Buffer[tempbuf[c]]^).rect.Left;
TGDIHookRec(pointer(dword(TextBuf) + sizeof(TGDIHookRec) * (c - 1))^).y:=
TPortion(Buffer[tempbuf[c]]^).rect.Top;
TGDIHookRec(pointer(dword(TextBuf) + sizeof(TGDIHookRec) * (c - 1))^).handle:=
TPortion(Buffer[tempbuf[c]]^).handle;
getmem(TGDIHookRec(pointer(dword(TextBuf) + sizeof(TGDIHookRec) * (c - 1))^).text,
lstrlenW(@TPortion(Buffer[tempbuf[c]]^).str[1]) * 2 +2);
lstrcpyW(TGDIHookRec(pointer(dword(TextBuf) + sizeof(TGDIHookRec) * (c - 1))^).text,
@TPortion(Buffer[tempbuf[c]]^).str[1]);
end;
except
log('Exception while preparing data');
end;
ReleaseMutex(ShareMutex);
{$ifdef debug}
log('Mutex released');
{$endif}
Elems := d;
{$ifdef debug}
log(format('GetTextFromWindow: elems = %d', [elems]));
{$endif}
result:=GDIHook_OK;
                      end;

Function PTInRect(rect:TRect; PT:TPoint):boolean;
begin
if (pt.X >= rect.Left) and (pt.X <= rect.Right) and (pt.Y >=rect.Top)
and (pt.Y <= rect.Bottom) then result:=true
else result:=false;
end;
Function OpenStorage:boolean;
begin
result:=false;
{$ifdef debug}
log('OpenStorage called');
{$endif}
try
Buffer:=TSharedArray.Create(StorageShareName, true, SizeOf(TPortion));
except
log('Exception while TSharedArray.create');
end;
    if not buffer.Success then exit;
ShareMutex:=CreateMutex(nil, false, ShareMutexName);
if ShareMutex = 0 then begin
log('OpenStorage: CreateMutex failed');
exit;
end;
result:=true;
end;
end.
